/**********************************************************************
 * config_file.c                                          November 1999
 * Horms                                             horms@vergenet.net
 *
 * Read in a config and parse it into command line arguments,
 * return this as a dynamic array
 *
 * Released under the terms of the GNU GPL
 * 
 **********************************************************************/


#include "config_file.h"

/**********************************************************************
 * config_file_to_opt
 * Configure opt structire according to options specified in a config
 * file.
 * pre: filename: file to read options from
 * post: options in global options_t opt are set according to
 *       config file. Options specified onthe command line
 *       override config file options
 **********************************************************************/

void config_file_to_opt(const char *filename){
  dynamic_array_t *a;

  if((a=config_file_read(filename))==NULL){
    return;
  }
  
  /* Set options according to config file but only if they are not 
   * Masked (overriden on the command line)
   */

  options(
    dynamic_array_get_count(a),
    dynamic_array_get_vector(a),
    OPT_USE_MASK|OPT_FILE
  );

  dynamic_array_destroy(a, DESTROY_STR);

  return;
}


/**********************************************************************
 * config_file_read
 * Read in a config file and put elements in a dynamic array
 * pre: filename: file to read configuration from
 * return: dynamic array containin elements, keys are preceded by
 *         a -- and must be long opts as per options.c.
 *         If a key is a single letter it is preceded by a -
 *         and must be a short opt as per options.c
 *         A key is the first whitespace delimited word on a line
 *         Blank lines are ignored.
 *         Everthing including and after a hash (#) on a line is 
 *         ignored
 **********************************************************************/

dynamic_array_t *config_file_read (const char *filename){
  char key[MAX_LINE_LENGTH+3];
  char tail[MAX_LINE_LENGTH+1];
  char format[MAX_LINE_LENGTH];
  int status;
  int token=0;
  int comment=0;
  char *s;
  dynamic_array_t *a;
  FILE *stream;

  extern int errno;
  extern options_t opt;

  if(filename==NULL) return(NULL);
  if((stream=fopen(filename, "r"))==NULL){
    if(opt.debug){
      syslog(
        LOG_WARNING, 
        "config_file_read: fopen(%s): %s", 
        filename,
        strerror(errno)
      );
    }
    return(NULL);
  }

  if((a=dynamic_array_create((size_t)0))==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "config_file_read: dynamic_array_create");
    }
    return(NULL);
  }
  /*insert a dummy argv[0] into the dynamic array*/
  if((a=dynamic_array_add_element(a, "", DESTROY_STR, DUP_STR))==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "config_file_read: dynamic_array_add_element");
    }
    return(NULL);
  }

  *key='-';
  *(key+1)='-';
  sprintf(
    format,
    "%%%d[^ \t\n\r#]%%%d[ \t\n\r#]",
    MAX_LINE_LENGTH, 
    MAX_LINE_LENGTH
  );
  while((status=fscanf(stream, format, key+2, tail))!=EOF){
    if(status==0) { status=fscanf(stream, "%[^\n]\n", tail); continue; }
    if(!comment){
      if((a=dynamic_array_add_element(
        a, 
        key+(token==0?((*(key+3)=='\0')?1:0):2), 
        DESTROY_STR,
        DUP_STR
      ))==NULL){
        if(opt.debug){
          syslog(LOG_WARNING, "config_file_read: dynamic_array_add_element");
        }
        return(NULL);
      }
    }
    if(status==2){
      if((s=strrchr(tail, '\n'))==NULL){
        token++;
      }
      else{
        token=0;
        comment=0;
      }
      if(!comment){
        comment=(strchr(s!=NULL?s:tail, '#')==NULL)?0:1;
      }
    }
  }
  return(a);
}


/**********************************************************************
 * config_file_reread_handler
 * A signal handler that rereads the config file on signal
 * and resets the signal handler
 * pre: sig: signal recieved by the process
 * post: config file reread
 *       signal handler reset for signal
 **********************************************************************/

void config_file_reread_handler(int sig){
  extern options_t opt;

  config_file_to_opt(opt.config_file);
  if(opt.debug){
    syslog(LOG_INFO, "Config file reread on signal (%d)\n", sig);
  }
  log_options();
  signal(sig, (void(*)(int))config_file_reread_handler);
  return;
}

